from elegy import types
from elegy.utils import Deferable
import typing as tp
import haiku as hk
import jax.numpy as jnp
from abc import abstractmethod
from elegy import utils


class Metric(hk.Module, Deferable):
    """
    Encapsulates metric logic and state.

    Usage:

    ```python
    m = SomeMetric(...)
    for input in ...:
        result = m(input)
    print('Final result: ', result)
    ```

    Usage with the Model API:

    ```python
    class MLP(elegy.Module):
        def call(self, image: jnp.ndarray) -> jnp.ndarray:
            mlp = hk.Sequential([
                hk.Flatten(),
                hk.Linear(300),
                jax.nn.relu,
                hk.Linear(10),
            ])
            return mlp(image)

    model = elegy.Model(
        module=MLP.defer(),
        loss=[
            elegy.losses.SparseCategoricalCrossentropy(from_logits=True),
        ],
        metrics=[
            elegy.metrics.SparseCategoricalAccuracy.defer()
        ],
        optimizer=optix.rmsprop(1e-3),
    )
    ```

    To be implemented by subclasses:

    * `call()`: All state variables should be created in this method by
        calling `haiku.get_state()`, update this state by calling
        `haiku.set_state(...)`, and return a result based on these states.

    Example subclass implementation:

    ```python
    class Accuracy(elegy.Metric):
        def call(self, y_true, y_pred):

            total = hk.get_state("total", [], jnp.zeros)
            count = hk.get_state("count", [], jnp.zeros)

            total += jnp.sum(y_true == y_pred)
            count += jnp.prod(y_true.shape)

            hk.set_state("total", total)
            hk.set_state("count", count)

            return total / count
    ```
    """

    def __init__(
        self,
        name: tp.Optional[str] = None,
        dtype: tp.Optional[jnp.dtype] = None,
        on: tp.Optional[types.IndexLike] = None,
    ):
        """
        Base Metric constructor.

        Arguments:
            name: string name of the metric instance.
            dtype: data type of the metric result.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """

        super().__init__(name=name)

        self._dtype = self._dtype = dtype if dtype is not None else jnp.float32
        self._labels_filter = (on,) if isinstance(on, (str, int)) else on
        self.call = utils.inject_dependencies(self.call)

    def __call__(self, y_true=None, y_pred=None, **kwargs):

        if self._labels_filter is not None:
            if y_true is not None:
                for index in self._labels_filter:
                    y_true = y_true[index]

            if y_pred is not None:
                for index in self._labels_filter:
                    y_pred = y_pred[index]

        return self.call(y_true=y_true, y_pred=y_pred, **kwargs)

    @abstractmethod
    def call(self, *args, **kwargs):
        ...



from elegy import types
import typing as tp

import jax.numpy as jnp

from elegy.losses.binary_crossentropy import binary_crossentropy
from elegy.metrics.mean import Mean


class BinaryCrossentropy(Mean):
    """
    Computes the crossentropy metric between the labels and predictions.
    This is the crossentropy metric class to be used when there are only two
    label classes (0 and 1).

    Usage:
    ```python
    y_true=jnp.array([[0., 1.], [0., 0.]]),
    y_pred=jnp.array([[0.6, 0.4], [0.4, 0.6]])
    
    bce = elegy.metrics.BinaryCrossentropy()
    result = bce(
        y_true=y_true,
        y_pred=y_pred,
    )
    assert jnp.isclose(result, 0.815, rtol=0.01)

    # BCE using sample_weight
    bce = elegy.metrics.BinaryCrossentropy()
    result = bce(y_true, y_pred, sample_weight=jnp.array([1., 0.]))
    assert jnp.isclose(result, 0.916, rtol=0.01)
    ```

    Usage with elegy API:

    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.CategoricalCrossentropy(),
        metrics=elegy.metrics.BinaryCrossentropy.defer(),
    )
    ```
    """

    def __init__(
        self,
        from_logits: bool = False,
        name: tp.Optional[str] = None,
        dtype: tp.Optional[jnp.dtype] = None,
        on: tp.Optional[types.IndexLike] = None,
    ):
        """
        Creates a `BinaryCrossentropy` instance.

        Arguments:
            from_logits: Whether `y_pred` is expected to be a logits tensor. By
                default, we assume that `y_pred` encodes a probability distribution.
                **Note - Using from_logits=True is more numerically stable.**
            name: string name of the metric instance.
            dtype: data type of the metric result.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """

        super().__init__(name=name, dtype=dtype, on=on)
        self._from_logits = from_logits

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:

        """
            Accumulates metric statistics. `y_true` and `y_pred` should have the same shape.
            
            Arguments:
                y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
                y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.
                sample_weight: Optional `sample_weight` acts as a
                    coefficient for the metric. If a scalar is provided, then the metric is
                    simply scaled by the given value. If `sample_weight` is a tensor of size
                    `[batch_size]`, then the metric for each sample of the batch is rescaled
                    by the corresponding element in the `sample_weight` vector. If the shape
                    of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be broadcasted
                    to this shape), then each metric element of `y_pred` is scaled by the
                    corresponding value of `sample_weight`. (Note on `dN-1`: all metric
                    functions reduce by 1 dimension, usually the last axis (-1)).
            Returns:
                Array with the cumulative accuracy.
        """

        return super().call(
            values=binary_crossentropy(
                y_true=y_true, y_pred=y_pred, from_logits=self._from_logits
            ),
            sample_weight=sample_weight,
        )



import haiku as hk
import jax.numpy as jnp
from haiku.testing import transform_and_run
from unittest import TestCase

import elegy


class BatchNormalizationTest(TestCase):
    @transform_and_run
    def test_connects(self):
        elegy.nn.BatchNormalization()(jnp.ones([3, 3]), is_training=True)

    def test_on_predict(self):
        class TestModule(elegy.Module):
            def call(self, x, is_training):
                return elegy.nn.BatchNormalization()(x, is_training)

        model = elegy.Model(module=TestModule.defer())

        x = jnp.ones([3, 5])

        y_pred = model.predict(x)
        logs = model.evaluate(x)



from elegy import types
import typing as tp

import jax.numpy as jnp

from elegy import utils
from elegy.losses.loss import Loss, Reduction


def mean_squared_error(y_true: jnp.ndarray, y_pred: jnp.ndarray) -> jnp.ndarray:
    """
    Computes the mean squared error between labels and predictions.
    
    After computing the squared distance between the inputs, the mean value over
    the last dimension is returned.
    
    ```python
    loss = mean(square(y_true - y_pred), axis=-1)
    ```

    Usage:
    
    ```python
    rng = jax.random.PRNGKey(42)

    y_true = jax.random.randint(rng, shape=(2, 3), minval=0, maxval=2)
    y_pred = jax.random.uniform(rng, shape=(2, 3))

    loss = elegy.losses.mean_squared_error(y_true, y_pred)

    assert loss.shape == (2,)

    assert jnp.array_equal(loss, jnp.mean(jnp.square(y_true - y_pred), axis=-1))
    ```
    
    Arguments:
        y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`.
        y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.
    
    Returns:
        Mean squared error values. shape = `[batch_size, d0, .. dN-1]`.
    """

    y_true = y_true.astype(y_pred.dtype)

    return jnp.mean(jnp.square(y_pred - y_true), axis=-1)


class MeanSquaredError(Loss):
    """
    Computes the mean of squares of errors between labels and predictions.

    `loss = square(y_true - y_pred)`

    Usage:

    ```python
    y_true = jnp.array([[0.0, 1.0], [0.0, 0.0]])
    y_pred = jnp.array([[1.0, 1.0], [1.0, 0.0]])

    # Using 'auto'/'sum_over_batch_size' reduction type.
    mse = elegy.losses.MeanSquaredError()

    assert mse(y_true, y_pred) == 0.5

    # Calling with 'sample_weight'.
    assert mse(y_true, y_pred, sample_weight=jnp.array([0.7, 0.3])) == 0.25

    # Using 'sum' reduction type.
    mse = elegy.losses.MeanSquaredError(reduction=elegy.losses.Reduction.SUM)

    assert mse(y_true, y_pred) == 1.0

    # Using 'none' reduction type.
    mse = elegy.losses.MeanSquaredError(reduction=elegy.losses.Reduction.NONE)

    assert list(mse(y_true, y_pred)) == [0.5, 0.5]
    ```
    Usage with the Elegy API:

    ```python
    model = elegy.Model(
        module_fn,
        loss=elegy.losses.MeanSquaredError(),
        metrics=elegy.metrics.Mean.defer(),
    )
    ```
    """

    def __init__(
        self,
        reduction: tp.Optional[Reduction] = None,
        name: tp.Optional[str] = None,
        weight: tp.Optional[float] = None,
        on: tp.Optional[types.IndexLike] = None,
    ):
        """
        Initializes `Mean` class.

        Arguments:
            reduction: (Optional) Type of `elegy.losses.Reduction` to apply to
                loss. Default value is `SUM_OVER_BATCH_SIZE`. For almost all cases
                this defaults to `SUM_OVER_BATCH_SIZE`.
            name: Optional name for the loss.
            weight: Optional weight contribution for the total loss. Defaults to `1`.
            on: A string or integer, or iterable of string or integers, that
                indicate how to index/filter the `y_true` and `y_pred`
                arguments before passing them to `call`. For example if `on = "a"` then
                `y_true = y_true["a"]`. If `on` is an iterable
                the structures will be indexed iteratively, for example if `on = ["a", 0, "b"]`
                then `y_true = y_true["a"][0]["b"]`, same for `y_pred`. For more information
                check out [Keras-like behavior](https://poets-ai.github.io/elegy/guides/modules-losses-metrics/#keras-like-behavior).
        """

        return super().__init__(reduction=reduction, name=name, weight=weight, on=on)

    def call(
        self,
        y_true: jnp.ndarray,
        y_pred: jnp.ndarray,
        sample_weight: tp.Optional[
            jnp.ndarray
        ] = None,  # not used, __call__ handles it, left for documentation purposes.
    ) -> jnp.ndarray:
        """
        Invokes the `MeanSquaredError` instance.

        Arguments:
            y_true: Ground truth values. shape = `[batch_size, d0, .. dN]`, except
                sparse loss functions such as sparse categorical crossentropy where
                shape = `[batch_size, d0, .. dN-1]`
            y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`
            sample_weight: Optional `sample_weight` acts as a
                coefficient for the loss. If a scalar is provided, then the loss is
                simply scaled by the given value. If `sample_weight` is a tensor of size
                `[batch_size]`, then the total loss for each sample of the batch is
                rescaled by the corresponding element in the `sample_weight` vector. If
                the shape of `sample_weight` is `[batch_size, d0, .. dN-1]` (or can be
                broadcasted to this shape), then each loss element of `y_pred` is scaled
                by the corresponding value of `sample_weight`. (Note on`dN-1`: all loss
                functions reduce by 1 dimension, usually axis=-1.)

        Returns:
            Weighted loss float `Tensor`. If `reduction` is `NONE`, this has
                shape `[batch_size, d0, .. dN-1]`; otherwise, it is scalar. (Note `dN-1`
                because all loss functions reduce by 1 dimension, usually axis=-1.)
        
        Raises:
            ValueError: If the shape of `sample_weight` is invalid.
        """
        return mean_squared_error(y_true, y_pred)



# Implementation based on Tensorflow Keras and Haiku
# Tensorflow: https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/keras/layers/normalization.py#L46
# Haiku: https://github.com/deepmind/dm-haiku/blob/master/haiku/_src/batch_norm.py#L39#L194

import typing as tp

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from haiku._src import utils as hk_utils

from elegy.module import Module


class BatchNormalization(Module):
    r"""
    Normalize and create_scale inputs or activations. (Ioffe and Szegedy, 2014).

    Normalize the activations of the previous layer at each batch,
    i.e. applies a transformation that maintains the mean activation
    close to 0 and the activation standard deviation close to 1.
    
    ### Normalization equations

    Consider the intermediate activations \(x\) of a mini-batch of size
    \\(m\\):
    We can compute the mean and variance of the batch
    
    \\({\mu_B} = \frac{1}{m} \sum_{i=1}^{m} {x_i}\\)
    \\({\sigma_B^2} = \frac{1}{m} \sum_{i=1}^{m} ({x_i} - {\mu_B})^2\\)
    
    and then compute a normalized \\(x\\), including a small factor
    \\({\eps}\\) for numerical stability.
    
    \\(\hat{x_i} = \frac{x_i - \mu_B}{\sqrt{\sigma_B^2 + \eps}}\\)
    
    And finally \\(\hat{x}\\) is linearly transformed by \\({\gamma}\\)
    and \\({\beta}\\), which are learned parameters:
    
    \\({y_i} = {\gamma * \hat{x_i} + \beta}\\)
    
    ### References

    * [Batch Normalization: Accelerating Deep Network Training by Reducing
        Internal Covariate Shift](https://arxiv.org/abs/1502.03167)
    """

    def __init__(
        self,
        axis: tp.Union[int, tp.Sequence[int], None] = None,
        momentum: float = 0.99,
        eps: float = 1e-3,
        create_offset: bool = True,
        create_scale: bool = True,
        scale_init: tp.Optional[hk.initializers.Initializer] = None,
        offset_init: tp.Optional[hk.initializers.Initializer] = None,
        cross_replica_axis: tp.Optional[str] = None,
        data_format: str = "channels_last",
        name: tp.Optional[str] = None,
    ):
        """
        Creates a BatchNormalization instance.
        
        Arguments:
            axis: Integer, the axis that should be normalized
                (typically the features axis).
                For instance, after a `Conv2D` layer with
                `data_format="channels_first"`,
                set `axis=1` in `BatchNormalization`.
            momentum: Momentum for the moving average.
            eps: Small float added to variance to avoid dividing by zero.
            create_offset: If True, add offset of `beta` to normalized tensor.
                se, `beta` is ignored.
            create_scale: If True, multiply by `gamma`.
                If False, `gamma` is not used.
                When the next layer is linear (also e.g. `relu`),
                this can be disabled since the scaling
                will be done by the next layer.
            scale_init: Optional initializer for gain (aka create_scale). Can only be set
                if ``create_scale=True``. By default, ``1``.
            offset_init: Optional initializer for bias (aka offset). Can only be set
                if ``create_offset=True``. By default, ``0``.
            axis: Which axes to reduce over. The default (``None``) signifies that all
                but the channel axis should be normalized. Otherwise this is a list of
                axis indices which will have normalization statistics calculated.
            cross_replica_axis: If not ``None``, it should be a string representing
                the axis name over which this module is being run within a ``jax.pmap``.
                Supplying this argument means that batch statistics are calculated
                across all replicas on that axis.
            data_format: The data format of the input. Can be either
                ``channels_first``, ``channels_last``, ``N...C`` or ``NC...``. By
                default it is ``channels_last``.
            name: The module name.
        """
        super().__init__(name=name)

        if not create_scale and scale_init is not None:
            raise ValueError("Cannot set `scale_init` if `create_scale=False`")
        if not create_offset and offset_init is not None:
            raise ValueError("Cannot set `offset_init` if `create_offset=False`")

        self.create_scale = create_scale
        self.create_offset = create_offset
        self.eps = eps
        self.scale_init = scale_init or jnp.ones
        self.offset_init = offset_init or jnp.zeros
        self.axis = axis
        self.cross_replica_axis = cross_replica_axis
        self.channel_index = hk_utils.get_channel_index(data_format)
        self.mean_ema = hk.ExponentialMovingAverage(momentum, name="mean_ema")
        self.var_ema = hk.ExponentialMovingAverage(momentum, name="var_ema")

    def call(
        self,
        inputs: jnp.ndarray,
        is_training: bool,
        test_local_stats: bool = False,
        create_scale: tp.Optional[jnp.ndarray] = None,
        offset: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        """Computes the normalized version of the input.

    Arguments:
        inputs: An array, where the data format is ``[..., C]``.
        is_training: Whether training is currently happening.
        test_local_stats: Whether local stats are used when is_training=False.
        create_scale: An array up to n-D. The shape of this tensor must be broadcastable
            to the shape of ``inputs``. This is the create_scale applied to the normalized
            inputs. This cannot be passed in if the module was constructed with
            ``create_scale=True``.
        offset: An array up to n-D. The shape of this tensor must be broadcastable
            to the shape of ``inputs``. This is the offset applied to the normalized
            inputs. This cannot be passed in if the module was constructed with
            ``create_offset=True``.

    Returns:
        The array, normalized across all but the last dimension.
    """
        if self.create_scale and create_scale is not None:
            raise ValueError(
                "Cannot pass `create_scale` at call time if `create_scale=True`."
            )
        if self.create_offset and offset is not None:
            raise ValueError(
                "Cannot pass `offset` at call time if `create_offset=True`."
            )

        channel_index = self.channel_index
        if channel_index < 0:
            channel_index += inputs.ndim

        if self.axis is not None:
            axis = self.axis
        else:
            axis = [i for i in range(inputs.ndim) if i != channel_index]

        if is_training or test_local_stats:
            cross_replica_axis = self.cross_replica_axis
            if self.cross_replica_axis:
                mean = jnp.mean(inputs, axis, keepdims=True)
                mean = jax.lax.pmean(mean, cross_replica_axis)
                mean_of_squares = jnp.mean(inputs ** 2, axis, keepdims=True)
                mean_of_squares = jax.lax.pmean(mean_of_squares, cross_replica_axis)
                var = mean_of_squares - mean ** 2
            else:
                mean = jnp.mean(inputs, axis, keepdims=True)
                # This uses E[(X - E[X])^2].
                # TODO(tycai): Consider the faster, but possibly less stable
                # E[X^2] - E[X]^2 method.
                var = jnp.var(inputs, axis, keepdims=True)
        else:
            mean = self.mean_ema.average
            var = self.var_ema.average

        if is_training:
            self.mean_ema(mean)
            self.var_ema(var)

        w_shape = [1 if i in axis else inputs.shape[i] for i in range(inputs.ndim)]
        w_dtype = inputs.dtype

        if self.create_scale:
            create_scale = hk.get_parameter(
                "create_scale", w_shape, w_dtype, self.scale_init
            )
        elif create_scale is None:
            create_scale = np.ones([], dtype=w_dtype)

        if self.create_offset:
            offset = hk.get_parameter("offset", w_shape, w_dtype, self.offset_init)
        elif offset is None:
            offset = np.zeros([], dtype=w_dtype)

        inv = create_scale * jax.lax.rsqrt(var + self.eps)
        return (inputs - mean) * inv + offset



# Implementation based on Tensorflow Keras and Haiku
# Tensorflow: https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/layers/core.py#L127-L224
# Haiku: https://github.com/deepmind/dm-haiku/blob/master/haiku/_src/basic.py#L293#L315

import typing as tp

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from elegy.module import Module


class Dropout(Module):
    """
    Applies Dropout to the input.

    The Dropout layer randomly sets input units to 0 with a frequency of `rate`
    at each step during training time, which helps prevent overfitting.
    Inputs not set to 0 are scaled up by `1/(1 - rate)` such that the sum over
    all inputs is unchanged.

    Note that the Dropout layer only applies when `is_training` is set to `True`
    such that no values are dropped during inference. When using `model.fit`,
    `is_training` will be appropriately set to True automatically, and in other
    contexts, you can set the kwarg explicitly to True when calling the layer.

    ### Example
    ```python
    dropout = elegy.nn.Dropout(0.2)
    data = np.arange(10).reshape(5, 2).astype(np.float32)
    
    print(data)
    # [[0. 1.]
    # [2. 3.]
    # [4. 5.]
    # [6. 7.]
    # [8. 9.]]

    outputs = dropout(data, is_training=True)
    
    print(outputs)
    # [[ 0.    1.25]
    # [ 2.5   3.75]
    # [ 5.    6.25]
    # [ 7.5   0.  ]
    # [10.    0.  ]]
    ```
    """

    def __init__(self, rate, name: tp.Optional[str] = None):
        super().__init__(name=name)
        self.rate = rate

    def call(
        self, x: np.ndarray, is_training: bool, rng: tp.Optional[np.ndarray] = None,
    ) -> jnp.ndarray:
        """
        Arguments:
            x: The value to be dropped out.
            is_training: Whether training is currently happening.
            rng: Optional RNGKey.
        Returns:
            x but dropped out and scaled by `1 / (1 - rate)`.
        """
        return hk.dropout(
            rng=hk.next_rng_key() if rng is None else rng,
            rate=self.rate if is_training else 0.0,
            x=x,
        )
